@@ -60,6 +60,9 @@
 #include <winsock2.h>
 #endif
 
+#ifdef __linux
+#define __USE_GNU
+#endif
 #include <fcntl.h>
 #include <sys/types.h>
 #ifdef HAVE_SYS_UIO_H
@@ -82,6 +85,8 @@
 #endif
 #endif /* HAVE_POLL_H */
 
+#include <sys/statvfs.h>
+
 #ifndef NOSOUND
 
 #ifdef HAVE_MACHINE_SOUNDCARD_H
@@ -144,9 +149,15 @@ volatile sig_atomic_t signal_report;
 volatile sig_atomic_t n_sighups;
 volatile sig_atomic_t childpid = 0;
 
+int pagesize = -1;
+
+struct pmalloc {
+  void *data;
+  void *allocated;
+  size_t n_bytes;
+};
 struct options {
   int v;
-  int b;
   long long n;
   int t;
   char *i;
@@ -160,9 +171,13 @@ struct options {
   int w;
   int S;
   int T;
+  int user_specified_blocksize;
 };
 
+void malloc_page_aligned(const struct options *o, struct pmalloc * pmalloc);
+
 struct progstate {
+  int b; /* Blocksize, to be changed to separate input and output sizes */
   int ifd;
   int ofd;
   long long bytes_transferred; /* How many bytes transferred so far */
@@ -179,6 +194,8 @@ struct progstate {
   int pid;
   int teefd; /* Usually -1, but can be a file descriptor number to copy
 	      * the stream to.  Currently that will always be 1 (stdout) */
+  int using_o_direct;
+  struct pmalloc pmalloc; /* Size is b (blocksize) */
 };
 
 #ifndef NOTIMER 
@@ -297,15 +314,20 @@ static void tsignal(const int sig, void (*const handler)(int))
 static void setaudio(int fd, const char *spec)
 {
   struct soundoptions {
-    int format;
-    int rate;
-    int stereo;
+    int so_format;
+    int so_rate;
+    int so_stereo;
   } so = {
     /* Default settings: CD quality */
     AFMT_S16_LE,
     44100,
     1
   };
+  char *s;
+
+  if ((s = getenv("CSTREAM_AUDIO_BITRATE"))) {
+    so.so_rate = atol(s);
+  }
 
   if (spec == NULL)
     spec = "-";
@@ -314,13 +336,13 @@ static void setaudio(int fd, const char *spec)
     fprintf(stderr, "Warning: audio options will just be CD-quality "
 	    "settings.\nFilespec parsing not implemented\n");
 
-  if (ioctl(fd, SNDCTL_DSP_SETFMT, &(so.format)) == -1)
+  if (ioctl(fd, SNDCTL_DSP_SETFMT, &(so.so_format)) == -1)
     fprintf(stderr, "icotl SNDCTL_DSP_SETFMT for '%s' failed: '%s'\n"
 	    , spec, strerror(errno));
-  if (ioctl(fd, SNDCTL_DSP_STEREO, &(so.stereo)) == -1)
+  if (ioctl(fd, SNDCTL_DSP_STEREO, &(so.so_stereo)) == -1)
     fprintf(stderr, "icotl SNDCTL_DSP_STEREO for '%s' failed: '%s'\n"
 	    , spec, strerror(errno));
-  if (ioctl(fd, SNDCTL_DSP_SPEED, &(so.rate)) == -1)
+  if (ioctl(fd, SNDCTL_DSP_SPEED, &(so.so_rate)) == -1)
     fprintf(stderr, "icotl SNDCTL_DSP_SPEED for '%s' failed: '%s'\n"
 	    , spec, strerror(errno));
 }
@@ -515,23 +537,82 @@ static int open_tcp(const struct options *const o, int mode)
 }
 #endif /* else NOTCP */
 
+void open_output_file(const struct options *const o
+		      , struct progstate *const state, int flags)
+{
+  int mode = 0666;
 
-static void init(struct options *const o, struct progstate *const state)
+  if (strchr(o->O, 'S')) {
+    flags |= O_SYNC;
+    if (o->v > 1) {
+      fprintf(stderr, "Using O_SYNC on output file\n");
+    }
+  }
+
+  if (state->using_o_direct) {
+    flags |= O_DIRECT;
+    if (o->v > 1) {
+      fprintf(stderr, "Using O_DIRECT on output file with blocksize %d\n"
+	      , state->b);
+    }
+
+    if (!o->user_specified_blocksize) {
+      struct statvfs statfs;
+
+      state->ofd = open(o->o, flags, mode);
+      if (state->ofd == -1) {
+	perror("Cannot open output file (to get block size)");
+	exit(2);
+      }
+      if (fstatvfs(state->ofd, &statfs) == -1) {
+	perror("Cannot get FS blocksize, you need to use -b<blocksize>\n");
+	exit(2);
+      }
+      state->b = statfs.f_bsize;
+      close(state->ofd);
+    }
+  }
+
+  state->ofd = open(o->o, flags, mode);
+  if (state->ofd == -1) {
+    perror("Cannot open output file");
+    exit(2);
+  }
+}
+
+static void init(struct options *const o, struct progstate *const state
+		 , const int blocksize)
 {
   struct timeval t;
 
-  if (o->b == 0)
-    o->b = 8192;
+  bzero(state, sizeof(*state));
 
-  if (o->B == 0)
-    o->B = o->b;
+  pagesize = getpagesize();
+
+  if (o->I == NULL)
+    o->I = "";
+  if (o->O == NULL)
+    o->O = "";
+
+  if (strchr(o->O, 'D'))
+    state->using_o_direct = 1;
+
+  if (blocksize == 0) {
+    state->b = 8192;
+  } else {
+    state->b = blocksize;
+    o->user_specified_blocksize = 1;
+  }
+
+  if (state->b == 0)
+    state->b = state->b;
 
   if (o->w == 0)
-    o->w = o->b;
+    o->w = state->b;
 
-  if (o->B < o->b) {
+  if (state->b < state->b) {
     fprintf(stderr, "-B must not be lower than -b or -n (%d/%d)\n"
-	    , o->B, o->b);
+	    , state->b, state->b);
     exit(1);
   }
 
@@ -540,11 +621,6 @@ static void init(struct options *const o, struct progstate *const state)
     exit(1);
   }
 
-  if (o->I == NULL)
-    o->I = "";
-  if (o->O == NULL)
-    o->O = "";
-
   if (o->i == NULL || o->i[0] == '\0')
     state->ifd = 0;
   else if (!strcmp(o->i, "-"))
@@ -597,11 +673,7 @@ static void init(struct options *const o, struct progstate *const state)
     if (strchr(o->o, ':') && !strchr(o->O, 'N'))
       state->ofd = open_tcp(o, O_WRONLY);
     else {
-      state->ofd = open(o->o, O_RDWR | O_CREAT | O_TRUNC, 0666);
-      if (state->ofd == -1) {
-	perror("Cannot open output file");
-	exit(2);
-      }
+      open_output_file(o, state, O_RDWR | O_CREAT | O_TRUNC); /* sets ofd */
     }
   }
 
@@ -675,6 +747,10 @@ static void init(struct options *const o, struct progstate *const state)
   }
   state->starttime = (double)t.tv_sec + (double)t.tv_usec / 1000000.0;
   state->lasttime = -1.0;
+
+
+  state->pmalloc.n_bytes = state->b;
+  malloc_page_aligned(o, &state->pmalloc);
 }
 
 static void closefiles(struct progstate *const state)
@@ -749,8 +825,8 @@ static void report(const struct options *const o,
   print_kmg(" ", "%.2f", rate, "B/s", stderr);
   if (o->l)
       fprintf(stderr, " %g lines", (double)state->n_lines);
-  if (curbytes != -1 && o->B != o->b)
-    fprintf(stderr, " %.1f %%buf", (double)curbytes / (double)o->B * 100.0);
+  if (curbytes != -1 && state->b != state->b)
+    fprintf(stderr, " %.1f %%buf", (double)curbytes / (double)state->b * 100.0);
   fprintf(stderr, "\n");
 
   if (o->v >= 2 && (sofar - state->time_firsttransfer) > 0.0) {
@@ -789,6 +865,82 @@ static void report(const struct options *const o,
   }
 }
 
+int my_write(const struct options *const o, struct progstate *const state
+	     , const void *buf, const size_t n_bytes) {
+  int ret;
+  static int warning_printed = 0;
+
+  if (state->using_o_direct && (long)buf % pagesize != 0) {
+    if (warning_printed == 0) {
+      warning_printed = 1;
+      fprintf(stderr, "Write from buffer not page aligned, copying data\n");
+    }
+    if (state->pmalloc.n_bytes < n_bytes) {
+      fprintf(stderr, "Target buffer too small, reallocating\n");
+      free(state->pmalloc.allocated);
+      state->pmalloc.n_bytes = n_bytes;
+      malloc_page_aligned(o, &state->pmalloc);
+    }
+    memcpy(state->pmalloc.data, buf, n_bytes);
+    buf = state->pmalloc.data;
+  }
+
+  ret = write(state->ofd, buf, n_bytes);
+  if (ret != -1)
+    return ret;
+
+  /* Error */
+
+  /* This is fine */
+  if (errno == EINTR)
+    return 0;
+
+  /* If we were doing O_DIRECT, retry after reopening */
+  if (state->using_o_direct) {
+    fprintf(stderr, "Write of %llu bytes failed at %p (%lld)"
+	    ", reopening without O_DIRECT\n"
+	    , (unsigned long long)n_bytes
+	    , buf, (long long)(long)buf % pagesize);
+    if (close(state->ofd) == -1) {
+      perror("Cannot close output file\n");
+      exit(2);
+    }
+    state->using_o_direct = 0;
+    open_output_file(o, state, O_RDWR | O_APPEND);
+  }
+
+  ret = write(state->ofd, buf, n_bytes);
+  if (ret != -1)
+    return ret;
+
+  if (errno == EINTR)
+    return 0;
+
+  report(o, state, -1);
+  perror("write");
+  exit(2);
+}
+
+void malloc_page_aligned(const struct options *const o
+			 , struct pmalloc *const pmalloc)
+{
+  pmalloc->allocated = malloc(pmalloc->n_bytes + pagesize);
+
+  if (pmalloc->allocated == NULL) {
+    perror("malloc");
+    exit(2);
+  }
+
+  pmalloc->data = pmalloc->allocated + pagesize 
+    - (long)pmalloc->allocated % pagesize;
+
+  if (o->v >=2) {
+    fprintf(stderr, "Page-aligned malloc at %p -> %p\n"
+	    , pmalloc->allocated, pmalloc->data);
+  }
+
+}
+
 #if HAVE_POLL_H
 static void pollloop(const struct options *const o
 		     , struct progstate *const state)
@@ -800,11 +952,11 @@ static void pollloop(const struct options *const o
   struct pollfd *pfd;
   int want_to_read;
   int want_to_write;
+  struct pmalloc palloc;
 
-  if ((buf = malloc(o->B)) == 0) {
-    perror("malloc");
-    exit(2);
-  }
+  palloc.n_bytes = state->b;
+  malloc_page_aligned(o, &palloc);
+  buf = palloc.data;
 
   pollfd[0].fd = state->ifd;
   pollfd[0].events = POLLIN|POLLRDNORM;
@@ -823,7 +975,7 @@ static void pollloop(const struct options *const o
     /* Space in buffer for another read */
     if (curread - buf /* Bytes in Buffer */
 	<=
-	o->B - o->b /* Max bytes in buffer for another read() */
+	state->b - state->b /* Max bytes in buffer for another read() */
 	) {
       want_to_read = 1;
     } else {
@@ -866,15 +1018,8 @@ static void pollloop(const struct options *const o
 	nbytes = curread - curwrite;
 
       if (o->v >= 4)
-	fprintf(stderr, "Trying to write %d bytes\n", nbytes);
-      ret = write(state->ofd, curwrite, nbytes);
-
-      if (ret == -1) {
-	if (errno == EINTR)
-	  continue;
-	perror("poll");
-	exit(2);
-      }
+	fprintf(stderr, "Trying to write %d bytes (1)\n", nbytes);
+      ret = my_write(o, state, curwrite, nbytes);
       if (o->v >= 4)
 	fprintf(stderr, "Wrote %d bytes\n", ret);
       curwrite += ret;
@@ -884,10 +1029,10 @@ static void pollloop(const struct options *const o
       if (o->v >= 4)
 	fprintf(stderr, "Trying to read\n");
 
-      if (o->n && state->bytes_transferred + o->b > o->n) {
+      if (o->n && state->bytes_transferred + state->b > o->n) {
 	ret = read(state->ifd, curread, o->n - state->bytes_transferred);
       } else
-	ret = read(state->ifd, curread, o->b);
+	ret = read(state->ifd, curread, state->b);
 
       if (ret == -1) {
 	if (errno == EINTR)
@@ -921,19 +1066,20 @@ static void pollloop(const struct options *const o
 
   closefiles(state);
 
-  free(buf);
+  free(palloc.allocated);
 }
 #endif /* HAVE_POLL_H */
 
 static void loop(struct options *const o, struct progstate *const state)
 {
-  char *buf; /* Holds the whole buffer o->B*/
+  char *buf; /* Holds the whole buffer state->b*/
   char *curbuf; /* Iterates over buf when reading and leave after last
 		 * read position */
   char *curbuf2; /* Iterate over buf when writing */
   int nbytes = 0; /* Just one system call read() */
   long long bytes_read = 0; /* Overall, both loops */
   int pipefd[2];
+  struct pmalloc palloc;
 
   state->pid = getpid();
   if (o->c > 0) {
@@ -970,19 +1116,17 @@ static void loop(struct options *const o, struct progstate *const state)
 
   if (o->c == 1) { /* Reader will buffer */
     if (childpid == 0) { /* Then writer will just use block size */
-      o->B = o->b;
+      state->b = state->b;
     }
   } else if (o->c == 2) {
     if (childpid != 0) { /* Otherwise reader will just use block size */
-      o->B = o->b;
+      state->b = state->b;
     }
   }
 
-  buf = malloc(o->B);
-  if (buf == NULL) {
-    perror("Cannot allocate memory");
-    exit(2);
-  }
+  palloc.n_bytes = state->b;
+  malloc_page_aligned(o, &palloc);
+  buf = palloc.data;
 
   if (o->c == 0 || childpid != 0) { /* Only in parent process */
     if (state->ifd == -1) {
@@ -995,7 +1139,7 @@ static void loop(struct options *const o, struct progstate *const state)
 	/* ASCII text for non-audio */
 	int i;
 
-	for (i = 0; i < o->B; i++) {
+	for (i = 0; i < state->b; i++) {
 	  /* Newlines below 80 chars and at end of buffer */
 	  if ((i+1) % 76 == 0 || i == o->n - 1)
 	    buf[i] = '\n';
@@ -1015,12 +1159,12 @@ static void loop(struct options *const o, struct progstate *const state)
 	int frames = rate / freq;
 	unsigned short val;
 
-	if (frames * 2 * 2 > o->b) {
+	if (frames * 2 * 2 > state->b) {
 	  fprintf(stderr, "Blocksize too small for wave\n");
 	  exit(1);
 	}
 
-	o->b = o->B = frames * 2 * 2; /* 16 bit Stereo */
+	state->b = state->b = frames * 2 * 2; /* 16 bit Stereo */
 	for (i = 0; i < frames; i++) {
 	  val = sin((double)i * 2.0 * pi / (double)frames) * 32767.0;
 	  buf[i * 4] = buf[i * 4 + 2] = val % 256;
@@ -1045,7 +1189,7 @@ static void loop(struct options *const o, struct progstate *const state)
   while (stopit == 0) {
     /* Outer loop - encapsulates pairs of read/write loops */
 
-    for (curbuf = buf; stopit == 0 && curbuf + o->b <= buf + o->B;
+    for (curbuf = buf; stopit == 0 && curbuf + state->b <= buf + state->b;
 	 curbuf += nbytes, bytes_read += nbytes) {
       /* Inner loop, fill buffer with reads() */
 
@@ -1085,13 +1229,13 @@ static void loop(struct options *const o, struct progstate *const state)
 
 	  /* fixme efficiency - can be moved to be computed once */
 	  /* how much time should it have taken? */
-	  theory = o->b / (double) - o->t;
+	  theory = state->b / (double) - o->t;
 
 	  time_to_sleep = (theory - sofar) * 1000000.0;
 	  if (o->v >= 4)
 	    fprintf(stderr,"We needed %g seconds for %d bytes"
 		    ", should be %g, sleeping %g\n"
-		    , sofar, o->b, theory, time_to_sleep);
+		    , sofar, state->b, theory, time_to_sleep);
 
 	  state->lasttime = ((double)t2.tv_sec +
 			     (double)t2.tv_usec / 1000000.0);
@@ -1107,7 +1251,7 @@ static void loop(struct options *const o, struct progstate *const state)
 
       if (state->ifd == -1) {
 	/* Just generate data */
-	if (o->n && bytes_read + o->b > o->n) {
+	if (o->n && bytes_read + state->b > o->n) {
 	  nbytes = o->n - bytes_read;
 	  if (o->v > 1)
 	    fprintf(stderr, "#%d finishing on condition 3\n",
@@ -1115,13 +1259,13 @@ static void loop(struct options *const o, struct progstate *const state)
 	  stopit = 1;
 	}
 	else
-	  nbytes = o->b;
+	  nbytes = state->b;
       } else {
 	if (o->v >= 4)
 	  fprintf(stderr,"#%d trying to read %d bytes\n", childpid ? 1 : 0,
-		  o->b);
+		  state->b);
 
-	if (o->n && bytes_read + o->b > o->n) {
+	if (o->n && bytes_read + state->b > o->n) {
 	  nbytes = read(state->ifd, curbuf, o->n - bytes_read);
 	  if (o->v > 1)
 	    fprintf(stderr, "#%d finishing on condition 4\n",
@@ -1129,7 +1273,7 @@ static void loop(struct options *const o, struct progstate *const state)
 	  stopit = 1;
 	}
 	else
-	  nbytes = read(state->ifd, curbuf, o->b);
+	  nbytes = read(state->ifd, curbuf, state->b);
 
 	if (o->v >= 4)
 	  fprintf(stderr,"#%d read %d bytes\n", childpid ? 1 : 0, nbytes);
@@ -1173,34 +1317,33 @@ static void loop(struct options *const o, struct progstate *const state)
     /* Write out */
     for (curbuf2 = buf; curbuf2 < curbuf; curbuf2 += nbytes) {
       if (state->ofd == -1) {
-	if (curbuf - curbuf2 < o->b)
+	if (curbuf - curbuf2 < state->b)
 	  nbytes = curbuf - curbuf2;
 	else
-	  nbytes = o->b;
+	  nbytes = state->b;
       } else {
-	if (curbuf - curbuf2 < o->b)
-	  nbytes = write(state->ofd, curbuf2, curbuf - curbuf2);
+	int n;
+
+	if (curbuf - curbuf2 < state->b)
+	  n = curbuf - curbuf2;
 	else
-	  nbytes = write(state->ofd, curbuf2, o->b);
-	if (nbytes == -1) {
-	  if (errno == EINTR) {
-	    nbytes = 0;
-	  } else {
-	    perror("write() failed");
-	    report(o, state, -1);
-	    exit(2);
-	  }
-	}
+	  n = state->b;
+
+	if (o->v >= 4)
+	  fprintf(stderr, "Trying to write %d bytes (2) from %p\n", n
+		  , curbuf2);
+	nbytes = my_write(o, state, curbuf2, n);
+
 	if (o->v >= 4)
 	  fprintf(stderr,"#%d wrote %d bytes\n", childpid ? 1 : 0, nbytes);
 
 	/* ? tee mode */
 	if ((o->c == 0 || childpid != 0) && /* Only in parent process */
 	    state->teefd != -1) {
-	  if (curbuf - curbuf2 < o->b)
+	  if (curbuf - curbuf2 < state->b)
 	    nbytes = write(state->teefd, curbuf2, curbuf - curbuf2);
 	  else
-	    nbytes = write(state->teefd, curbuf2, o->b);
+	    nbytes = write(state->teefd, curbuf2, state->b);
 	  if (nbytes == -1) {
 	    if (errno == EINTR) {
 	      nbytes = 0;
@@ -1256,7 +1399,7 @@ static void loop(struct options *const o, struct progstate *const state)
     }
   }
 
-  free(buf);
+  free(palloc.allocated);
 }
 
 void cleanup(const struct options *const o, struct progstate *const state)
@@ -1306,6 +1449,8 @@ static void usage(void)
           "         'N' = don't use TCP even if filename has ':'\n"
           "         't' = tee - in addition to outfile, "
 	  "copy stream to fd 3\n"
+          "         'D' = O_DIRECT\n"
+          "         'S' = O_SYNC\n"
 	  "         [Multiple chars allowed]\n"
 	  "-p <s> = Write pid as ascii integer to file <s>\n"
 	  "-l       include line count in statistics\n"
@@ -1342,26 +1487,14 @@ int main(int argc, char *const argv[])
   struct options o;
   struct progstate state;
   int ch;
+  size_t blocksize = 0;
 
   default_options(&o);
 
   while ((ch = getopt(argc, argv, "b:B:c:i:I:n:o:O:p:St:T:v:Vl")) != -1) {
     switch(ch) {
-    case 'v': {
-      if (optarg[1] != '\0' ||
-	  (optarg[0] != '0' &&
-	   optarg[0] != '1' &&
-	   optarg[0] != '2' &&
-	   optarg[0] != '3' &&
-	   optarg[0] != '4')) {
-	fprintf(stderr, "Error: option -v requires a decimal numeric argument "
-		"between 0 and 4.\nType -? for usage instructions\n");
-	exit(1);
-      }
-      o.v = atoi(optarg);
-      break;
-    }
-    case 'b': o.b = atoi_kmg(optarg); break;
+    case 'v': o.v = atoi(optarg); break;
+    case 'b': blocksize = atoi_kmg(optarg); break;
     case 'B': o.B = atoi_kmg(optarg); break;
     case 'c': o.c = atoi(optarg); break;
     case 'l': o.l = 1; break;
@@ -1399,7 +1532,7 @@ int main(int argc, char *const argv[])
   default: usage();
   }
 
-  init(&o, &state);
+  init(&o, &state, blocksize);
 
   if (o.T) {
 #ifndef NOTIMER 
@@ -1427,7 +1560,8 @@ int main(int argc, char *const argv[])
 
   cleanup(&o, &state);
 
-  /* Freeing of memory omitted */
+  free(state.pmalloc.allocated);
+  /* More freeing of memory omitted */
 
   return 0;
 }
